function [y, time] = chirp(startFreq, endFreq, samplingRate, numSeconds, method)
%CHIRP generates an ascending chirp signal.
%   Y = CHIRP(startFreq, endFreq, samplingRate, numSeconds)
%   where startFreq and endFreq (endFreq > startFreq) are the starting and
%   ending frequencies specified in Hz, samplingRate is the sampling rate
%   in Hz of the signal to be generated, and numSeconds is the duration of
%   the signal in seconds, returns a linear frequency modulated complex-
%   valued chirp signal.
%
%   Y = CHIRP(startFreq, endFreq, samplingRate, numSeconds, method)
%   specifies alternate chirp methods.
%   Available methods are 'linear' and 'hyperbolic'. The default is
%   'linear'.
%
%   [Y, TIME] = CHIRP(...) returns a vector of time indices in seconds
%   (TIME).

    if (nargin < 4)
        error('Too few arguments specified in function.');
    end
    
    if (nargin > 5)
        error('Too many arguments specified in function.');
    end    

    if (nargin == 4)
        method = 'linear';
    end

    time = 1/samplingRate : 1/samplingRate : numSeconds;
    if (strcmp(method, 'linear') == 1)
        beta = (endFreq - startFreq) / numSeconds;
        y = exp(1j * (2 * pi * (0.5 * beta * time.^2 + startFreq * time)));
    elseif (strcmp(method, 'hyperbolic') == 1)
        k = (endFreq - startFreq) / (endFreq * numSeconds);
        a = -2 * pi * startFreq / k;
        y = exp(1j * (a * log(1 - k * time)));        
    else
        error('Unknown chirp type %s.', method);
    end