//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

#include "Phy.h"
#include "Datagram.h"
#include "libphy.h"

Define_Module(Phy);

Phy::Phy() { }

Phy::~Phy() { }

void Phy::initialize() {
    modulationType = par("modulationType");
    if (strncmp(modulationType, "FSK", 3) == 0) {
        fskReceiverType = par("fskReceiverType");
        fskUseHardLimiter = par("fskUseHardLimiter");
    }
    samplingRate = par("samplingRate");
    carrierFrequency = par("carrierFrequency");
    symbolsPerSecond = par("symbolsPerSecond");
    signalLevelDB = par("signalLevelDB");
}

void Phy::handleMessage(cMessage *msg) {
	cGate* gate = msg->getArrivalGate();
	EV << "Arrived at phy layer on " << gate->getBaseName() << endl;
	Datagram *datagram = check_and_cast<Datagram *>(msg);
	if (datagram->arrivedOn("inFromLink")) {
	    // Modulation
	    datagram->setModulationType(modulationType);
	    datagram->setSamplingRate(samplingRate);
	    datagram->setCarrierFrequency(carrierFrequency);
        datagram->setSymbolsPerSecond(symbolsPerSecond);
        datagram->setSignalLevelDB(signalLevelDB);
        mxArray *x1_ptr,
                *x2_ptr,
                *x3_ptr,
                *x4_ptr;
        mxArray *y1_ptr = NULL,
                *y2_ptr = NULL;

        int byteStreamLength = datagram->getByteStreamLength();
        unsigned char *byteStream = datagram->getByteStream();
        const mwSize dims1[] = {1, byteStreamLength};
        x1_ptr = mxCreateNumericArray(2, dims1, mxUINT8_CLASS, mxREAL);
        unsigned char *startPtr = (unsigned char *)mxGetData(x1_ptr);
        size_t bytesToCopy = byteStreamLength * mxGetElementSize(x1_ptr);
        memcpy(startPtr, byteStream, bytesToCopy);

        x2_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getSamplingRate()));
        x3_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getCarrierFrequency()));
        x4_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getSymbolsPerSecond()));

        // Call the implementation function for modulation
        if (datagram->getModulationType() == Datagram::MOD_FSK) {
            mlfModulateFSK(2, &y1_ptr, &y2_ptr, x1_ptr, x2_ptr, x3_ptr, x4_ptr);
        } else {
            mlfModulatePSK(2, &y1_ptr, &y2_ptr, x1_ptr, x2_ptr, x3_ptr, x4_ptr);
        }
        double temp;
        memcpy(&temp, mxGetPr(y2_ptr), sizeof(double));
        int numberOfSamples = static_cast<int>(temp);
        cout << modulationType << "-modulated data in " << numberOfSamples << " samples." << endl;
        datagram->setModulatedSamples(numberOfSamples, mxGetPr(y1_ptr));

        mxDestroyArray(x1_ptr); x1_ptr = NULL;
        mxDestroyArray(x2_ptr); x2_ptr = NULL;
        mxDestroyArray(x3_ptr); x3_ptr = NULL;
        mxDestroyArray(x4_ptr); x4_ptr = NULL;
        mxDestroyArray(y1_ptr); y1_ptr = NULL;
        mxDestroyArray(y2_ptr); y2_ptr = NULL;

		send(datagram, "channelGate$o");
	} else {
	    // Demodulation
        mxArray *x1_ptr,
                *x2_ptr,
                *x3_ptr,
                *x4_ptr,
                *x5_ptr,
                *x6_ptr,
                *x7_ptr;
        mxArray *y1_ptr = NULL;

        int numberOfSamples = datagram->getConvolvedSamplesLength();
        double *samples = datagram->getConvolvedSamples();
        const mwSize dims1[] = {1, numberOfSamples};
        x1_ptr = mxCreateNumericArray(2, dims1, mxDOUBLE_CLASS, mxREAL);
        double *startPtr = (double *)mxGetData(x1_ptr);
        size_t bytesToCopy = numberOfSamples * mxGetElementSize(x1_ptr);
        memcpy(startPtr, samples, bytesToCopy);

        int numberOfBytes = datagram->getLength();
        cout << "Demodulating " << numberOfBytes << " bytes." << endl;
        x2_ptr = mxCreateDoubleScalar(static_cast<double>(numberOfBytes));
        x3_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getSamplingRate()));
        x4_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getCarrierFrequency()));
        x5_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getSymbolsPerSecond()));

        if (datagram->getModulationType() == Datagram::MOD_FSK) {
            x6_ptr = mxCreateString(fskReceiverType);
            x7_ptr = mxCreateDoubleScalar(static_cast<double>(fskUseHardLimiter));
            mlfDemodulateFSK(1, &y1_ptr, x1_ptr, x2_ptr, x3_ptr, x4_ptr, x5_ptr, x6_ptr, x7_ptr);
            mxDestroyArray(x6_ptr); x6_ptr = NULL;
            mxDestroyArray(x7_ptr); x7_ptr = NULL;
        } else {
            mlfDemodulatePSK(1, &y1_ptr, x1_ptr, x2_ptr, x3_ptr, x4_ptr, x5_ptr);
        }
        datagram->setDemodulatedByteStream(numberOfBytes, mxGetPr(y1_ptr));
        double ber = datagram->computeBER();
        printf("BER: %.2f%%\n", ber);

        mxDestroyArray(x1_ptr); x1_ptr = NULL;
        mxDestroyArray(x2_ptr); x2_ptr = NULL;
        mxDestroyArray(x3_ptr); x3_ptr = NULL;
        mxDestroyArray(x4_ptr); x4_ptr = NULL;
        mxDestroyArray(x5_ptr); x5_ptr = NULL;
        mxDestroyArray(y1_ptr); y1_ptr = NULL;

		sendDelayed(datagram, datagram->getTransmissionDelayInSeconds(), "outToLink");
	}
}
