//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

#include "Datagram.h"
#include "CRC16.h"
#include <omnetpp.h>
#include <string.h>

Register_Class(Datagram);

Datagram::Datagram(const cPacket &packet): cPacket(packet) {
	modulationTypeStr = NULL;
	byteStream = NULL;
	modulatedSamples = NULL;
	demodulatedByteStream = NULL;
	convolvedSamples = NULL;
	numberOfBytes = headerLengthInBytes;
}

Datagram::Datagram(const char *name, short kind, int64 length): cPacket(name, kind, length) {
    modulationTypeStr = NULL;
    byteStream = NULL;
    modulatedSamples = NULL;
    demodulatedByteStream = NULL;
    convolvedSamples = NULL;
    numberOfBytes = headerLengthInBytes;
}

Datagram::~Datagram() {
    if (modulationTypeStr != NULL) {
        delete modulationTypeStr;
        modulationTypeStr = NULL;
    }
    if (byteStream != NULL) {
        delete byteStream;
        byteStream = NULL;
    }
    if (modulatedSamples != NULL) {
        delete modulatedSamples;
        modulatedSamples = NULL;
    }
    if (demodulatedByteStream != NULL) {
        delete demodulatedByteStream;
        demodulatedByteStream = NULL;
    }
    if (convolvedSamples != NULL) {
        delete convolvedSamples;
        convolvedSamples = NULL;
    }
}

void Datagram::setPayload(string payload) {
    this->payload = payload;
    numberOfBytes += payload.length();
}

void Datagram::setSourceAddress(unsigned short sourceAddress) {
	this->sourceAddress = sourceAddress;
}

void Datagram::setDestinationAddress(unsigned short destinationAddress) {
	this->destinationAddress = destinationAddress;
}

void Datagram::setMessageID(int messageID) {
    this->messageID = messageID;
}

void Datagram::setModulationType(const char* modulationTypeStr) {
    this->modulationTypeStr = strdup(modulationTypeStr);
}

void Datagram::setSamplingRate(int samplingRate) {
    this->samplingRate = samplingRate;
}

void Datagram::setCarrierFrequency(int carrierFrequency) {
    this->carrierFrequency = carrierFrequency;
}

void Datagram::setSymbolsPerSecond(int symbolsPerSecond) {
    this->symbolsPerSecond = symbolsPerSecond;
}
void Datagram::setSignalLevelDB(double signalLevelDB) {
    this->signalLevelDB = signalLevelDB;
}

void Datagram::setChecksum(unsigned short checksum) {
    byteStream[0] = static_cast<unsigned char>((checksum & 0xFF00) >> 8);
    byteStream[1] = static_cast<unsigned char>(checksum & 0xFF);
}

void Datagram::setModulatedSamples(int length, double *samples) {
    numberOfModulatedSamples = length;
    modulatedSamples = new double[numberOfModulatedSamples];
    memcpy(modulatedSamples, samples, numberOfModulatedSamples * sizeof(double));
}

void Datagram::setDemodulatedByteStream(int length, double *bytes) {
    numberOfDemodulatedBytes = length;
    demodulatedByteStream = new unsigned char[numberOfDemodulatedBytes];
    for (int i = 0; i < numberOfDemodulatedBytes; i++) {
        demodulatedByteStream[i] = static_cast<unsigned char>(bytes[i]);
    }
}

void Datagram::setConvolvedSamples(int length, double *samples) {
    numberOfConvolvedSamples = length;
    convolvedSamples = new double[numberOfConvolvedSamples];
    memcpy(convolvedSamples, samples, numberOfConvolvedSamples * sizeof(double));
}

int Datagram::getLength() {
    return numberOfBytes;
}

string Datagram::getPayload() {
	return payload;
}

unsigned short Datagram::getSourceAddress() {
	return sourceAddress;
}

unsigned short Datagram::getDestinationAddress() {
	return destinationAddress;
}

int Datagram::getMessageID() {
    return messageID;
}

int Datagram::getModulationType() {
    if (strncmp(modulationTypeStr, "FSK", 3) == 0) {
        return MOD_FSK;
    } else {
        return MOD_PSK;
    }
}

const char* Datagram::getModulationTypeStr() {
    return modulationTypeStr;
}

int Datagram::getSamplingRate() {
    return samplingRate;
}

int Datagram::getCarrierFrequency() {
    return carrierFrequency;
}

int Datagram::getSymbolsPerSecond() {
    return symbolsPerSecond;
}

double Datagram::getTransmissionDelayInSeconds() {
    return static_cast<double>(numberOfBytes) * 8 / symbolsPerSecond;
}

double Datagram::getSignalLevelDB() {
    return signalLevelDB;
}

unsigned short Datagram::getChecksum() {
    return ((byteStream[0] << 8 | (byteStream[1] & 0xFF)) & 0x0000FFFF);
}

unsigned char* Datagram::getByteStream() {
    return byteStream;
}

int Datagram::getByteStreamLength() {
    if (byteStream == NULL) {
        return 0;
    } else {
        return payload.length() + headerLengthInBytes;
    }
}

double* Datagram::getModulatedSamples() {
    return modulatedSamples;
}

int Datagram::getModulatedSamplesLength() {
    return numberOfModulatedSamples;
}

unsigned char* Datagram::getDemodulatedByteStream() {
    return demodulatedByteStream;
}

int Datagram::getDemodulatedByteStreamLength() {
    return numberOfDemodulatedBytes;
}

double* Datagram::getConvolvedSamples() {
    return convolvedSamples;
}

int Datagram::getConvolvedSamplesLength() {
    return numberOfConvolvedSamples;
}

unsigned short Datagram::computeChecksum() {
    if (byteStream == NULL) return 0;
    int length = payload.length() + headerLengthInBytes;
    byteStream[0] = byteStream[1] = 0;
    return CRC16::getChecksum(byteStream, length);
}

double Datagram::computeBER() {
    int bitErrors = 0;
    int i;
    int length = payload.length() + headerLengthInBytes;
    int actualValue, receivedValue;
    for (i = 0; i < length; i++) {
        char actual = byteStream[i],
             received = demodulatedByteStream[i];
        for (int j = 0x80; j > 0; j >>= 1) {
            if ((j & actual) == 0) {
                actualValue = 0;
            } else {
                actualValue = 1;
            }
            if ((j & received) == 0) {
                receivedValue = 0;
            } else {
                receivedValue = 1;
            }
            if (actualValue != receivedValue) {
                bitErrors++;
            }
        }
    }
    return static_cast<double>(bitErrors)/(length * 8) * 100;
}

void Datagram::constructByteStream() {
    int payloadLength = payload.length();
    int datagramLength = payloadLength + headerLengthInBytes;
    byteStream = new unsigned char[datagramLength];
    byteStream[0] = byteStream[1] = 0;
    byteStream[2] = static_cast<unsigned char>((datagramLength & 0xFF00) >> 8);
    byteStream[3] = static_cast<unsigned char>(datagramLength & 0xFF);
    byteStream[4] = static_cast<unsigned char>((sourceAddress & 0xFF00) >> 8);
    byteStream[5] = static_cast<unsigned char>(sourceAddress & 0xFF);
    byteStream[6] = static_cast<unsigned char>((destinationAddress & 0xFF00) >> 8);
    byteStream[7] = static_cast<unsigned char>(destinationAddress & 0xFF);
    const char* payloadStr = payload.c_str();
    for (int i = 0; i < payloadLength; i++) {
        byteStream[i + headerLengthInBytes] = static_cast<unsigned char>(payloadStr[i]);
    }
}

/**
 * Displays the bit representation of a message for easy debugging.
 */
void Datagram::displayMessage(bool useExpectedResults) {
    unsigned char *stream;
    int i, length = payload.length() + headerLengthInBytes;

    if (useExpectedResults) {
        stream = byteStream;
    } else {
        stream = demodulatedByteStream;
    }
    for (i = 0; i < length; i++) {
        char c = stream[i];
        for (int j = 0x80; j > 0; j >>= 1) {
            if ((j & c) == 0) {
                cout << "0";
            } else {
                cout << "1";
            }
        }
        if ((i + 1) % 8 == 0) {
            cout << endl;
        } else {
            cout << ":";
        }
    }
    if (i % 8 != 0) {
        cout << endl;
    }
}

void Datagram::displayDemodulatedMessage() {
    if (!demodulatedByteStream) return;
    for (int i = headerLengthInBytes; i < numberOfDemodulatedBytes; i++) {
        unsigned char c = demodulatedByteStream[i];
        if (isprint(c)) {
            cout << c;
        } else {
            cout << '.';
        }
    }
    cout << endl;
}
