//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

#include "UWChannel.h"
#include "Datagram.h"
#include "libchannel.h"
#include <omnetpp.h>
#include <stdlib.h>

Define_Channel(UWChannel);

UWChannel::UWChannel() { }

UWChannel::~UWChannel() { }

void UWChannel::initialize() {
    linkDistanceInMeters = par("linkDistanceInMeters");
    useAmbientNoise = par("useAmbientNoise");

    mxArray *x1_ptr;
    mxArray *y1_ptr = NULL;

    x1_ptr = mxCreateDoubleScalar(linkDistanceInMeters);

    // Call the implementation function for average sound velocity
    mlfGetAverageSoundVelocity(1, &y1_ptr, x1_ptr);
    memcpy(&soundVelocityInMetersPerSecond, mxGetPr(y1_ptr), sizeof(double));
    printf("Average sound velocity : %.2f m/s\n", soundVelocityInMetersPerSecond);

    mxDestroyArray(x1_ptr); x1_ptr = NULL;
    mxDestroyArray(y1_ptr); y1_ptr = NULL;

    propagationDelayInSeconds = linkDistanceInMeters/soundVelocityInMetersPerSecond;
}

bool UWChannel::deliver(cMessage *msg, simtime_t t) {
    Datagram *datagram = check_and_cast<Datagram *>(msg);
    mxArray *x1_ptr,
            *x2_ptr,
            *x3_ptr,
            *x4_ptr,
            *x5_ptr;
    mxArray *y1_ptr = NULL,
            *y2_ptr = NULL;
    double sourceLevelDB, transmissionLossDB, noiseLevelDB, snrDB;

    sourceLevelDB = datagram->getSignalLevelDB();

    x1_ptr = mxCreateDoubleScalar(linkDistanceInMeters);
    printf("Source level           : %.2f dB\n", sourceLevelDB);

    // Call the implementation function for transmission loss
    mlfGetTransmissionLossDB(1, &y1_ptr, x1_ptr);
    transmissionLossDB = mxGetPr(y1_ptr)[0];
    printf("Transmission loss      : %.2f dB\n", transmissionLossDB);

    mxDestroyArray(x1_ptr); x1_ptr = NULL;
    mxDestroyArray(y1_ptr); y1_ptr = NULL;

    x1_ptr = mxCreateString("data/Noise.csv");
    x2_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getCarrierFrequency()));
    x3_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getSymbolsPerSecond()));
    x4_ptr = mxCreateString("FSK");
    x5_ptr = mxCreateDoubleScalar(static_cast<double>(useAmbientNoise));

    // Call the implementation function for noise level
    mlfGetNoiseLevelDB(1, &y1_ptr, x1_ptr, x2_ptr, x3_ptr, x4_ptr, x5_ptr);
    noiseLevelDB = mxGetPr(y1_ptr)[0];
    printf("Noise level            : %.2f dB\n", noiseLevelDB);

    snrDB = sourceLevelDB - transmissionLossDB - noiseLevelDB;
    printf("SNR                    : %.2f dB\n", snrDB);

    mxDestroyArray(x1_ptr); x1_ptr = NULL;
    mxDestroyArray(x2_ptr); x2_ptr = NULL;
    mxDestroyArray(x3_ptr); x3_ptr = NULL;
    mxDestroyArray(x4_ptr); x4_ptr = NULL;
    mxDestroyArray(x5_ptr); x5_ptr = NULL;
    mxDestroyArray(y1_ptr); y1_ptr = NULL;

    int impulseResponseNumber = intrand(601);
    char impulseResponseFileName[128];
    sprintf(impulseResponseFileName, "data/IR_%dm/IR_%d.wav", static_cast<int>(linkDistanceInMeters), impulseResponseNumber);
    cout << "Opening IR file " << impulseResponseFileName << "." << endl;

    int numberOfSamples = datagram->getModulatedSamplesLength();
    double *samples = datagram->getModulatedSamples();
    const mwSize dims1[] = {1, numberOfSamples};
    x1_ptr = mxCreateNumericArray(2, dims1, mxDOUBLE_CLASS, mxREAL);
    double *startPtr = (double *)mxGetData(x1_ptr);
    size_t bytesToCopy = numberOfSamples * mxGetElementSize(x1_ptr);
    memcpy(startPtr, samples, bytesToCopy);

    x2_ptr = mxCreateString(impulseResponseFileName);
    x3_ptr = mxCreateDoubleScalar(snrDB);
    x4_ptr = mxCreateDoubleScalar(static_cast<double>(datagram->getSamplingRate()));

    mlfSimulateChannel(2, &y1_ptr, &y2_ptr, x1_ptr, x2_ptr, x3_ptr, x4_ptr);
    double temp;
    memcpy(&temp, mxGetPr(y2_ptr), sizeof(double));
    numberOfSamples = static_cast<int>(temp);
    cout << "Number of samples in packet after convolution: " << numberOfSamples << endl;
    datagram->setConvolvedSamples(numberOfSamples, mxGetPr(y1_ptr));

    t += propagationDelayInSeconds;

    // FIXME: this is not reusable this way in custom channels, put it into a base class function with the next line (levy)
    // i.e use template method...?
    EVCB.messageSendHop(datagram, getSourceGate(), propagationDelayInSeconds, SIMTIME_ZERO);

    // Hand datagram over to next gate
    cGate *nextgate = getSourceGate()->getNextGate();
    return nextgate->deliver(datagram, t);
}
